# Standard Libraries
import os
from collections.abc import Mapping
from typing import Any
import time
from functools import wraps
from logging import getLogger

from sys import (
    modules as sys_modules,
    argv as sys_argv
)

# Third Party Libraries
from fastapi import FastAPI


logger = getLogger(__name__)


def update_dict(base_dict: dict[Any, Any], new_dict: dict[Any, Any]) -> dict[Any, Any]:
    """Recursively update a base dictionary with values from a new dictionary."""
    for key, value in new_dict.items():
        if isinstance(value, Mapping):
            base_dict[key] = update_dict(base_dict.get(key, {}), value)
        else:
            base_dict[key] = value
    return base_dict


def get_main_script_name() -> str:
    """Returns the name of the main script being executed, handling indirect calls."""
    # The first item in sys.argv is always the path to the main script.
    main_script_path: str = sys_argv[0]
    # Extract the file name without the extension
    main_script_name = os.path.splitext(os.path.basename(main_script_path))[0]
    return main_script_name


def get_main_script_dir() -> str:
    main_script_path: str = sys_modules['__main__'].__file__
    main_script_dir: str = os.path.dirname(os.path.abspath(main_script_path))
    return main_script_dir


def wrap_message(message: str, symbol: str ='-', max_length: int = 130) -> str:
    # Split the message into lines and strip each line
    lines: list[str] = message.split('\n')

    # Ensure that each line does not exceed the maximum length
    adjusted_lines: list[str] = []
    for line in lines:
        # Break the line into segments that fit the max_length
        while len(line) > max_length:
            adjusted_lines.append(line[:max_length])
            line = line[max_length:]
        adjusted_lines.append(line)

    # Use the maximum length for the wrap length
    dashed_line: str = symbol * max_length
    final_message: str = f"{dashed_line}\n" + "\n".join(adjusted_lines) + '\n' + f"{dashed_line}"

    return final_message


def get_default_log_path() -> str:
    # Get the main script name without its directory path and extension
    script_name: str = os.path.splitext(os.path.basename(sys_argv[0]))[0]

    # Determine the common log directory based on the OS environment
    log_dir: str = os.path.join(os.path.expanduser('~'), 'logs')

    # Ensure the log directory exists
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    # Construct the full path for the log file
    log_file_path: str = os.path.join(log_dir, f"{script_name}.log")

    return log_file_path


def list_routes(app: FastAPI) -> list[str]:
    route_list: list = []
    for route in app.routes:
        route_list.append({
            "path": route.path,
            "name": route.name,
            "methods": route.methods
        })
    return route_list


def retry(max_retries: int = 3, delay_sec: int = 2, exceptions: tuple = (ConnectionError,)):
    """
    A decorator that adds retry logic to functions which may throw specified exceptions.
    
    Args:
    max_retries (int): The maximum number of retries before giving up.
    delay_sec (int): The delay between retries in seconds.
    exceptions (tuple): A tuple of exception classes to catch and retry.

    Returns:
    Any: The return value of the decorated function, if successful.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempts = 0
            while attempts < max_retries:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    attempts += 1
                    if attempts == max_retries:
                        logger.error(f"Maximum retry attempts reached for {func.__name__}")
                        raise
                    logger.info(f"Retrying {func.__name__} due to {e}, attempt {attempts}")
                    time.sleep(delay_sec)
        return wrapper
    return decorator
