# Standard Libraries
from abc import ABC, abstractmethod
from typing import Any

# Project Specific Libraries
from univocalcore.schemas.tasks.enums import Management, Status


class TrackerInterface(ABC):
    @abstractmethod
    def create_task(self, task_data: dict) -> str:
        """
        Creates a new task based on the provided data. This involves creating entries for both the task's status and its informational details.
        
        :param task_data: A dictionary containing all necessary information to define the task.
        :return: A Str that represent the task ID.
        """
        pass

    @abstractmethod
    def delete_task(self, task_id: str) -> None:
        """
        Delete a task based on the task Id. This involves deleting the info and status entries.
        
        :param task_id: The UUID Str of the task whose need to be deleted.
        :return: None.
        """
        pass

    @abstractmethod
    def get_task_status(self, task_id: str) -> Status:
        """
        Retrieves the current status of the specified task.
        
        :param task_id: The UUID Str of the task whose status is being queried.
        :return: An enum value from TaskStatus indicating the current status of the task.
        """
        pass

    @abstractmethod
    def update_task_status(self, task_id: str, status: str) -> bool:
        """
        Updates the status of an existing task to a new value.
        
        :param task_id: The UUID Str of the task to update.
        :param status: The new status from the TaskStatus enum to apply to the task.
        :return: false or true.
        """
        pass

    @abstractmethod
    def get_task_info(self, task_id: str) -> Any:
        """
        Retrieves detailed information about a specific task.
        
        :param task_id: The UUID Str of the task for which information is being retrieved.
        :return: A dictionary or object containing the task's detailed information.
        """
        pass


class TaskQueueInterface(ABC):
    """
    Abstract base class for a task queue system. This interface provides methods
    for publishing and consuming tasks from a message queue.
    """

    @abstractmethod
    def publish_task(self, topic: Management.TASKS, task_id: str) -> None:
        """
        Publishes a task to a specified topic in the task queue system.

        Parameters:
            topic (Tasks): The topic (or queue) to which the task will be published.
                           Uses an enum to ensure only predefined topics are used.
            task_id (str): The unique identifier of the task to be published.

        Returns:
            None
        """
        pass

    @abstractmethod
    def consume_task(self, topic: Management.TASKS) -> str:
        """
        Consumes a task from a specified topic.

        Parameters:
            topic (Tasks): The topic (or queue) from which the task will be consumed.
                           Uses an enum to ensure only predefined topics are used.

        Returns:
            str: A task ID.
        """
        pass


    
