# Standard Libraries
from logging import getLogger
from uuid import uuid4

# Third Party Libraries
from redis import Redis
from redis.exceptions import ResponseError as redisResponseError

# Project Specific Libraries
from univocalcore.taskmanagment.interfaces import TaskQueueInterface
from univocalcore.taskmanagment.models import RedisConfig
from univocalcore.schemas.tasks.enums import Management


logger = getLogger(__name__)


class RedisStreamTaskQueue(TaskQueueInterface):
    def __init__(self, consumer_group: str | None = None, **kwargs):
        """
        Initialize the Redis connection using the RedisQueueConfig.
        """
        config = RedisConfig(**kwargs)
        self.redis_client = Redis(
            host=config.host,
            port=config.port,
            # db=config.db,  # Assuming there's a db attribute in RedisQueueConfig
            # password=config.password,
            decode_responses=config.decode_responses
        )
        logger.info(f"RedisTaskQueue initialized with Redis server at {config.host}:{config.port}")

        self.consumer_group = consumer_group
        self.consumer_name = None

        if consumer_group:
            logger.info('Initilazing a new consumer.')
            self.consumer_name = self._generate_consumer_name(consumer_group)
            self._create_consumer_group(consumer_group)

    def _generate_consumer_name(self, consumer_group: str) -> str:
        """
        Generates a unique consumer name using the consumer group name and UUID.

        Parameters:
            consumer_group (str): The name of the consumer group.

        Returns:
            str: A unique consumer name.
        """
        name = f"{consumer_group}_{uuid4().hex}"
        logger.info(f"New Consumer: {name}.")
        return name

    def _create_consumer_group(self, consumer_group: str) -> None:
        """
        Creates a consumer group if it does not already exist.

        Parameters:
            consumer_group (str): The name of the consumer group.

        Returns:
            None
        """
        try:
            # Assuming '0' is the start point for the consumer group
            self.redis_client.xgroup_create(Management.TASKS.value, consumer_group, id='0', mkstream=True)
            logger.info(f"Created Redis Group: {consumer_group} successfully.")
        except redisResponseError as e:
            if "BUSYGROUP Consumer Group name already exists" not in str(e):
                logger.error(f"Failed to create group: {consumer_group}, {e}")
                raise

    def publish_task(self, topic: Management.TASKS, task_id: str) -> None:
        """
        Publishes a task to a specified topic in the Redis Streams task queue system.

        Parameters:
            topic (Management.TASKS): The topic (or queue) to which the task will be published.
            task_id (str): The unique identifier of the task to be published.

        Returns:
            None
        """
        self.redis_client.xadd(topic.value, {'task_id': task_id})

    def consume_task(self, topic: Management.TASKS) -> str:
        """
        Consumes a task from a specified topic in the Redis Streams task queue system using consumer groups.

        Parameters:
            topic (Management.TASKS): The topic (or queue) from which the task will be consumed.

        Returns:
            str: A task ID.
        """
        if not self.consumer_group or not self.consumer_name:
            raise ValueError("Consumer group and consumer name must be initialized for consuming tasks.")

        stream_data = self.redis_client.xreadgroup(self.consumer_group, self.consumer_name, {topic.value: '>'}, count=1, block=0)
        if stream_data:
            message_id, task_data = stream_data[0][1][0]
            self.redis_client.xack(topic.value, self.consumer_group, message_id)  # Acknowledge the message
            return task_data['task_id'].decode('utf-8')
        return None
    

if __name__ == "__main__":
    tracker = RedisStreamTaskQueue()