# Standard Libraries
import json
from logging import getLogger
from typing import Any
from uuid import uuid4

# Third Party Libraries
from redis import Redis, RedisError

# Project Specific Libraries
from univocalcore.taskmanagment.interfaces import TrackerInterface
from univocalcore.taskmanagment.models import RedisConfig
from univocalcore.schemas.tasks.enums import Status
from univocalcore.helpers import retry


logger = getLogger(__name__)


class RedisTracker(TrackerInterface):
    """
    This class provides functionality to manage tasks within a Redis data store. It implements the TrackerInterface,
    allowing for the creation, deletion, and updating of task statuses, as well as the retrieval of task statuses and
    information. The class uses a Redis client to interact with a Redis database, where tasks are tracked by unique
    identifiers (UUIDs). Each task has associated status and informational entries within Redis hash structures.

    Attributes:
        redis_base_key (str): Base key under which task data is stored in Redis.
        redis_status_key (str): Key under which task statuses are stored, derived from the base key.
        redis_info_key (str): Key under which task information is stored, derived from the base key.
        redis_client (Redis): A Redis client instance configured to connect to a Redis server.

    Methods:
        __init__(self, **kwargs): Initializes a RedisTracker instance with a Redis configuration.
        create_task(self, task_data: dict) -> str: Creates a new task entry in Redis and returns the task ID.
        update_task_status(self, task_id: str, status: Status) -> bool: Updates the status of a specified task.
        delete_task(self, task_id: str) -> None: Deletes all entries associated with a specified task in Redis.
        get_task_status(self, task_id: str) -> Status: Retrieves the status of a specified task.
        get_task_info(self, task_id: str) -> Any: Retrieves detailed information about a specified task.
        __create_info_entry(self, task_id: str, task_data: dict) -> None: Creates an informational entry for a task in Redis.

    The class also includes error handling for Redis-specific and connection errors, ensuring robust interaction
    with the Redis server.
    """
    redis_base_key = 'tasks'
    redis_status_key = f"{redis_base_key}:status"
    redis_info_key = f"{redis_base_key}:info"

    def __init__(self, **kwargs):
        config = RedisConfig(**kwargs)  # Validate and create a RedisConfig object
        self.redis_client = Redis(
            host=config.host,
            port=config.port,
            # db=config.db,
            # password=config.password,
            decode_responses=config.decode_responses
        )
        logger.info(f"RedisTaskTracker initialized with Redis server at {config.host}:{config.port}")


    def create_task(self, task_data: dict) -> str:
        """
        Creates a new task in the Redis data store and assigns it a unique identifier (UUID). This function handles the
        process of initializing both the informational and status entries for the task within Redis. If the status update fails,
        the task entries are removed to ensure consistency.

        Parameters:
            task_data (dict): A dictionary containing all the necessary data to define and describe the task. This data
                            includes task-specific parameters such as task type, associated user details, and other
                            metadata required to execute the task.

        Returns:
            str: A string representing the unique task ID (UUID) assigned to the newly created task.

        Raises:
            ConnectionError: If there is a failure in connecting to the Redis server during the task creation process.
            RedisError: If there is an issue during the Redis operations, such as failing to set the hash value.

        Example:
            >>> task_data = {
                'task_type': 'data_processing',
                'user_id': '12345',
                'parameters': {'file_path': '/data/input.csv'}
            }
            >>> task_id = tracker.create_task(task_data)
            >>> print(task_id)
            16fd2706-8baf-433b-82eb-8c7fada847da

        This function logs the creation process and outcome, aiding in debugging and tracing the lifecycle of task creation.
        """
        task_id = str(uuid4())
        logger.info(f"New Task ID - {task_id}")
        logger.info(f"Creating Info and Status Entries for - {task_id}")

        # Create Info Entry for the Job
        self.__create_info_entry(task_id, task_data)

        # Try to update 
        updated = self.update_task_status(task_id, Status.CREATED)
        
        if not updated:
            self.delete_task(task_id)

        logger.info(f"Successfully Created - {task_id}")
        return task_id


    def update_task_status(self, task_id: str, status: Status) -> bool:
        """
        Updates the status of a specified task in the Redis data store. This method sets the status value in a Redis hash
        where each task is identified by its UUID. The function logs the success or failure of the update operation, which
        can aid in debugging and maintaining the integrity of task status tracking.

        Parameters:
            task_id (str): The UUID of the task whose status needs to be updated. This should be a valid UUID string that
                        uniquely identifies a task in the system.
            status (Status): The new status to assign to the task. This status should be one of the predefined states in
                        the Status enumeration.

        Returns:
            bool: True if the status update was successful, False otherwise. A return value of False typically indicates
                a failure in Redis operations or a connection issue.

        Raises:
            ConnectionError: If there is an issue connecting to the Redis server during the update process.
            RedisError: If an error occurs during the Redis operation, such as if the Redis service is unavailable.

        Example:
            >>> task_id = '16fd2706-8baf-433b-82eb-8c7fada847da'
            >>> success = tracker.update_task_status(task_id, 'COMPLETED')
            >>> print(success)
            True

        This method is crucial for maintaining the lifecycle of tasks and ensuring that their current statuses are accurately
        reflected and logged within the system.
        """
        try:
            self.redis_client.hset(self.redis_status_key, task_id, status.value)
            logger.debug(f"Status Entry updated successfully, Status: {status.value} - {task_id}")
        except (ConnectionError, RedisError) as e:
            logger.error(f"Failed during task creation: {e}")
            return False
        
        return True
        

    @retry(max_retries=3, delay_sec=60, exceptions=(ConnectionError, RedisError))
    def delete_task(self, task_id: str) -> bool:
        """
        Deletes all entries associated with a specified task in the Redis data store. This method attempts to remove both the
        information and status entries from their respective Redis hashes and returns a boolean indicating whether any entries
        were actually deleted.

        Parameters:
            task_id (str): The UUID of the task to be deleted. This must be a valid UUID string that corresponds to an
                        existing task within the Redis hashes managed by this tracker.

        Returns:
            bool: True if any entries were deleted, False if no entries were found to delete.

        Raises:
            ConnectionError: If a connection issue occurs during the delete operation.
            RedisError: If a Redis-specific error occurs, preventing the task deletion.
            Exception: Catches and re-raises any general exceptions that might occur during the deletion process.

        Example:
            >>> task_id = '16fd2706-8baf-433b-82eb-8c7fada847da'
            >>> deleted = tracker.delete_task(task_id)
            >>> print(deleted)
            True  # If deletion is successful and entries were found and deleted

        Decorators:
            @retry(max_retries=3, delay_sec=60, exceptions=(ConnectionError, RedisError))
            This decorator applies a retry mechanism to the method, specifying that the method should be retried up to 3 times
            with a 60-second delay between retries, specifically for ConnectionError and RedisError exceptions.

        Note:
            This method logs both the initiation and completion of the delete operation, providing traceability and aiding in
            diagnostics in case of issues.
        """
        logger.info(f"Attempting to delete task {task_id}")
        try:
            # Attempt to delete the keys from Redis and check if they were actually deleted
            info_deleted = self.redis_client.hdel(self.redis_info_key, task_id)
            status_deleted = self.redis_client.hdel(self.redis_status_key, task_id)
            was_deleted = info_deleted > 0 or status_deleted > 0

        except Exception as e:
            logger.error(f"Error deleting task {task_id}: {e}")
            raise

        logger.info(f"Deletion of task {task_id} {'successful' if was_deleted else 'failed, no entries found'}")
        return was_deleted


    def get_task_status(self, task_id: str) -> Status:
        """
        Retrieves the current status of a specified task from the Redis data store. This method accesses the task's status
        using the task's UUID in a Redis hash. If the task is not found, it raises an exception.

        Parameters:
            task_id (str): The UUID of the task whose status is being queried. This must be a valid UUID string that uniquely
                        identifies a task in the system.

        Returns:
            Status: An enum value representing the current status of the task. The status is one of the predefined states in
                    the Status enumeration.

        Raises:
            ConnectionError: If there is a connection issue with the Redis server during the retrieval process.
            RedisError: If a Redis-specific error occurs, such as a failure to fetch the hash entry.
            KeyError: If no entry exists for the given task ID, indicating that the task does not exist or has been deleted.

        Example:
            >>> task_id = '16fd2706-8baf-433b-82eb-8c7fada847da'
            >>> status = tracker.get_task_status(task_id)
            >>> print(status)
            Completed

        This method is essential for monitoring the progress and current state of tasks within the system, providing vital
        information for task management and troubleshooting.
        """
        try:
            status = self.redis_client.hget(self.redis_status_key, task_id)
            if status is None:
                raise KeyError(f"Task ID {task_id} not found")
            logger.debug(f"Retrieved status for task {task_id}: {status}")
            return Status(status).value
        except (ConnectionError, RedisError) as e:
            logger.error(f"Failed to retrieve task status for {task_id}: {e}")
            raise
        except KeyError as e:
            logger.error(e)
            raise


    def get_task_info(self, task_id: str) -> Any:
        """
        Retrieves detailed information about a specific task by querying its UUID in the Redis data store. The information
        is stored in JSON format and is decoded into a dictionary or relevant data structure upon retrieval. This method
        ensures that detailed task information is accessible for further processing or monitoring purposes.

        Parameters:
            task_id (str): The UUID of the task for which information is being retrieved. This must be a valid UUID string
                        that uniquely identifies a task in the system.

        Returns:
            Any: A dictionary or an object containing detailed information about the task. The exact structure of this data
                depends on how the task was initially stored in Redis.

        Raises:
            ConnectionError: If there is an issue connecting to the Redis server during the information retrieval process.
            RedisError: If a Redis-specific error occurs, such as a failure to fetch the hash entry.
            KeyError: If no entry exists for the given task ID, indicating that the task does not exist or has been deleted.
            json.JSONDecodeError: If there is an error decoding the task information from JSON format.

        Example:
            >>> task_id = '16fd2706-8baf-433b-82eb-8c7fada847da'
            >>> task_info = tracker.get_task_info(task_id)
            >>> print(task_info)
            {'task_type': 'data_processing', 'user_id': '12345', 'parameters': {'file_path': '/data/input.csv'}}

        This method is crucial for obtaining comprehensive details about tasks, facilitating task management and debugging.
        """
        try:
            task_info_json = self.redis_client.hget(self.redis_info_key, task_id)
            if task_info_json is None:
                raise KeyError(f"Task ID {task_id} not found")
            task_info = json.loads(task_info_json)
            logger.debug(f"Retrieved info for task {task_id}: {task_info}")
            return task_info
        except (ConnectionError, RedisError) as e:
            logger.error(f"Failed to retrieve task info for {task_id}: {e}")
            raise
        except KeyError as e:
            logger.error(e)
            raise
        except json.JSONDecodeError as e:
            logger.error(f"Failed to decode JSON for task {task_id}: {e}")
            raise
    

    ## Helpers Methods ##
    def __create_info_entry(self, task_id: str, task_data: dict) -> None:
        """
        Creates an informational entry for a task in the Redis data store. This method serializes the task data to JSON
        format and stores it in a Redis hash under a unique task ID. It is designed to be used internally within the class
        to manage the storage of task details comprehensively.

        Parameters:
            task_id (str): The UUID of the task for which the information is being stored. This should be a valid UUID string
                        that uniquely identifies the task.
            task_data (dict): A dictionary containing all necessary information about the task that needs to be stored. The
                            data must be serializable into JSON format.

        Raises:
            TypeError: If the task data contains non-serializable values and cannot be converted to JSON.
            ValueError: If the JSON serialization fails, indicating that the task data is not properly formatted or contains
                        invalid data.
            ConnectionError: If there is an issue connecting to the Redis server during the data storage process.
            RedisError: If a Redis-specific error occurs, such as a failure to set the hash entry.

        Example:
            >>> task_id = '16fd2706-8baf-433b-82eb-8c7fada847da'
            >>> task_data = {'task_type': 'data_processing', 'details': {'user_id': '12345', 'description': 'Process data files'}}
            >>> tracker.__create_info_entry(task_id, task_data)
            # Successful creation logs the debug message and does not return a value.

        Note:
            This is a private method and should not be called directly from outside the class. It is intended to support
            other public methods within the class, such as `create_task`, by handling the detailed information storage.
        """
        # Try to serilaze the dict to a json.
        try:
            task_data_json = json.dumps(task_data)
        except TypeError as e:
            # This catches errors in json serialization, such as when the data is not serializable
            message = f"Failed to serialize task data: {e}"
            logger.exception(message)
            raise ValueError(message)

        # Try Set key to value within a redis hash
        try: 
            self.redis_client.hset(self.redis_info_key, task_id, task_data_json)
            logger.debug(f"Info Entry created successfully - {task_id}")

        except ConnectionError as e:
            message = f"Failed to connect to Redis: {e}"
            logger.exception(message)
            raise ConnectionError(message)
        
        except RedisError as e:
            message = f"Redis error occurred: {e}"
            logger.exception(message)
            raise RedisError(message) 


if __name__ == "__main__":
    tracker = RedisTracker()